"""
test the Gaussian Processes in bayesvalidrox
Class GaussianProcessSklearn
    build_meta_model       - x
    check_is_gaussian      no need (always true)
    build_kernels
    transform_x
    fit
    adaptive_regression    - x
    scale_x                - x 
    eval_meta_model        - x
To-Dos: 
what are other errors that need to be catched for the fit and eval_meta_model functions?
"""

import numpy as np
import pytest
import sys

sys.path.append("../src/")

from bayesvalidrox import GPESkl, Input
from bayesvalidrox.surrogate_models.input_space import InputSpace

@pytest.fixture
def GPE(): 
    
    input = Input()
    input.add_marginals()
    input.Marginals[0].dist_type = 'normal'
    input.Marginals[0].parameters = [0, 1]
    samples = np.array([[0.2], [0.8]])
    MetaModel = GPESkl(input)
    MetaModel.CollocationPoints = samples
    MetaModel.InputSpace = InputSpace(MetaModel.input_obj, MetaModel.meta_model_type)

    return MetaModel

def test_meta_model():
    """Construct GPSklearn without input"""
    input = Input()
    input.add_marginals()
    input.Marginals[0].dist_type = 'normal'
    input.Marginals[0].parameters = [0, 1]
    GPESkl(input)

def test_build_metamodel(GPE) -> None:
    """
    Build GPE
    """
    MetaModel = GPE
    MetaModel.CollocationPoints = np.array([[0.2], [0.8]])
    MetaModel.build_metamodel()

def test_add_input_space(GPE) -> None:
    """
    Add input space
    """
    MetaModel = GPE
    MetaModel.add_InputSpace()

def test_fit() -> None: 
    """
    Fit GPE
    """
    input = Input()
    input.add_marginals()
    input.Marginals[0].dist_type = 'normal'
    input.Marginals[0].parameters = [0, 1]
    MetaModel = GPESkl(input)
    MetaModel.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})

def test_fit_parallel() -> None:
    """
    Fit GPE in parallel
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    mm = GPESkl(inp)
    mm.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]}, parallel=True)

def test_fit_verbose() -> None:
    """
    Fit PCE verbose
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    mm = GPESkl(inp)
    mm.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]}, verbose=True)

def test_build_kernels(GPE) -> None:
    """
    Build kernels
    """
    MetaModel = GPE
    MetaModel.build_kernels()

def test_kernel_no_kertyp(GPE) -> None:
    """
    Test kernel with no kernel type
    """
    MetaModel = GPE
    MetaModel.build_kernels()
    assert MetaModel._kernel_type == 'RBF'

def test_kernel_wrong_kertyp(GPE) -> None:
    """
    Test kernel with no kernel type
    """
    MetaModel = GPE
    MetaModel.build_kernels()
    MetaModel.kernel_type = 123
    assert MetaModel._kernel_type == 'RBF'

def test_kernel_type_rbf(GPE) -> None:
    """
    Test kernel type
    """
    MetaModel = GPE
    MetaModel.kernel_type = 'rbf'
    MetaModel.build_kernels()

def test_build_kernels_matern(GPE) -> None:
    """
    Build kernels with Matern kernel type
    """
    MetaModel = GPE
    MetaModel.kernel_type = 'matern'
    MetaModel.build_kernels()

def test_build_kernels_rq(GPE) -> None:
    """
    Build kernels with Matern kernel type
    """
    MetaModel = GPE
    MetaModel.kernel_type = 'rq'
    MetaModel.build_kernels()


def test_build_kernels_with_length_scale(GPE) -> None:
    """
    Build kernels with specific length scale
    """
    MetaModel = GPE
    MetaModel.kernel_length_scale = 1.5
    MetaModel.build_kernels()

def test_build_kernels_with_bounds(GPE) -> None:
    """
    Build kernels with specific bounds
    """
    MetaModel = GPE
    MetaModel.kernel_bounds = (1e-2, 1e1)
    MetaModel.build_kernels()


def test_kernel_isotropy(GPE):
    """
    Test kernel isotropy
    """
    MetaModel = GPE
    MetaModel.kernel_isotropy = True
    MetaModel.build_kernels()

def test_adaptive_regression(GPE) -> None: 
    """
    Regression without a specific method 
    """
    MetaModel = GPE
    samples = np.array([[0.0], [0.1]])
    outputs = np.array([[0.1], [0.1]])

    MetaModel.build_metamodel()

    MetaModel.adaptive_regression(samples, outputs, varIdx=None)

def test_adaptive_regression_verbose(GPE) -> None: 
    """
    Regression without a specific method with verbose
    """
    MetaModel = GPE
    samples = np.array([[0.0], [0.1]])
    outputs = np.array([[0.1], [0.1]])

    MetaModel.build_metamodel()

    MetaModel.adaptive_regression(samples, outputs, varIdx=None, verbose=True)

#TO Do: 
def test_scale_x(GPE):
    """
    Scale x
    """
    MetaModel = GPE
    x_s, scaler = MetaModel.transform_x(np.array([[0.1], [0.2]]))
    MetaModel.scale_x(np.array([[0.1], [0.2]]), scaler)

def test_eval_metamodel () -> None: 
    """
    Evaluate trained GPE
    """
    input = Input()
    input.add_marginals()
    input.Marginals[0].dist_type = 'normal'
    input.Marginals[0].parameters = [0, 1]
    MetaModel = GPESkl(input)
    MetaModel.out_names = ['Z']
    MetaModel.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})
    MetaModel.eval_metamodel([[0.4]])

def test_eval_metamodel_multiple_outputs() -> None:
    """
    Evaluate trained GPE with multiple outputs
    """
    input = Input()
    input.add_marginals()
    input.Marginals[0].dist_type = 'normal'
    input.Marginals[0].parameters = [0, 1]
    MetaModel = GPESkl(input)
    MetaModel.out_names = ['Z1', 'Z2']
    MetaModel.fit([[0.2], [0.4], [0.8]], {'Z1': [[0.4], [0.2], [0.5]], 'Z2': [[0.3], [0.1], [0.4]]})
    MetaModel.eval_metamodel([[0.4]])

